/*
    Hurry Curry! - a game about cooking
    Copyright (C) 2026 Hurry Curry! Contributors

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, version 3 of the License only.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/
use super::simple::State;
use crate::{BotAlgo, PacketSink, algos::simple::Context, step::StepState};
use hurrycurry_game_core::Game;
use hurrycurry_protocol::{ItemIndex, PlayerID};

pub struct DishWasher {
    me: PlayerID,
    step: StepState,
    dirty_plate: Option<ItemIndex>,
}

type LogicRes<Out = ()> = Result<Out, ()>;

impl DishWasher {
    pub fn new(me: PlayerID) -> Self {
        Self {
            me,
            step: StepState::new_idle(me, "idle".to_string()),
            dirty_plate: None,
        }
    }
}
impl BotAlgo for DishWasher {
    fn tick(&mut self, mut out: PacketSink, game: &Game, dt: f32) {
        if self.dirty_plate.is_none() {
            self.dirty_plate = game.data.get_item_by_name("dirty-plate");
            if self.dirty_plate.is_none() {
                return;
            }
        }
        if self.step.is_busy() {
            return self.step.tick(&mut out, game, dt);
        }

        Context {
            game,
            me: self.me,
            state: self,
            complexity_abort: 0,
            stack: Vec::new(),
        }
        .update()
        .ok();
    }
}

impl State for DishWasher {
    fn queue_step(&mut self, step: StepState) {
        self.step = step;
    }
}

impl Context<'_, DishWasher> {
    fn update(&mut self) -> LogicRes {
        if let Some(sink) = self.find_free_tool_by_name("sink") {
            if self.is_hand_item(self.state.dirty_plate.unwrap()) {
                // TODO clear sink first but dont "steal" items from others working there
                self.interact_with(sink, 2.0, "wash")?;
            } else {
                self.assert_hand_is_clear()?;
            }
            if let Some(pos) = self.find_item_on_map(false, self.state.dirty_plate.unwrap()) {
                self.interact_with(pos, 0., "pick dirty plate")?;
            }
            self.wait(0.5, "idle")?;
        } else {
            self.assert_hand_is_clear()?;
        }
        Ok(())
    }
}
